import { parseArgs } from "@std/cli";
import { blue, bold, yellow } from "@std/fmt/colors";
import { emptyDirSync, ensureDirSync, existsSync } from "@std/fs";
import { faker } from "https://deno.land/x/deno_faker@v1.0.3/mod.ts";
import {
  CreatureUtil,
  CRISPR,
  Mutation,
  NeatOptions,
  Selection,
} from "@stsoftware/neat-ai";
import { addTag, getTag, removeTag } from "@stsoftware/tags";
import { NetworkUtil } from "./NetworkUtil.ts";
import { TRAINING_DATA_DIR as BASE_TRAINING_DATA_DIR } from "./train/Constants.ts";
import { CrisprInterface } from "@stsoftware/neat-ai";

const args = parseArgs(Deno.args);

let mutation = Mutation.ALL;
const mutationName = args.mutation
  ? args.mutation
  : Math.random() > 0.5
  ? "ALL"
  : "FFW";
switch (mutationName) {
  case "FFW":
    mutation = Mutation.FFW;
    break;
  case "ALL":
    break;
  default:
    console.error("Unknown mutation: " + mutationName);
    Deno.exit(1);
}

function editAliases(fileName: string): CrisprInterface {
  const crispr: CrisprInterface = JSON.parse(
    Deno.readTextFileSync(`CRISPR/${fileName}`),
  );
  const aliases = JSON.parse(Deno.readTextFileSync("src/v/Aliases.json"));
  return CRISPR.editAliases(crispr, aliases);
}

const TRAINING_DATA_DIR = BASE_TRAINING_DATA_DIR +
  (args.sampler ? "-sampler" : "");
const dir = args.dir ? args.dir : ".creatures";
ensureDirSync(dir);
const nu = new NetworkUtil();
let networkFN = ".network.json";
if (args.island) {
  networkFN = `.islands/${args.island}.json`;
  // mutation = [
  //   Mutation.SUB_NODE,
  //   Mutation.SUB_CONN,
  //   Mutation.MOD_WEIGHT,
  //   Mutation.MOD_BIAS,
  //   Mutation.MOD_ACTIVATION,
  // ];
  // mutationName = mutation.map((m: { name: string }) => m.name).join(",");
  if (!existsSync(networkFN)) {
    Deno.copyFileSync(`islands/${args.island}.json`, networkFN);
  }
} else if (args.sampler == "true") {
  networkFN = ".network-sampler.json";
  if (existsSync(".network-sampler.json")) {
    if (existsSync(".network.json")) {
      Deno.copyFileSync(".network.json", `${dir}/network.json`);
    }
  }
}

const creature = await nu.load(networkFN);

const scoreTxt = getTag(creature, "score");
const name = getTag(creature, "name");

const startScore = scoreTxt ? parseFloat(scoreTxt) : -1;
console.info(
  `${bold(blue(name ? name : "John Doe"))} ` +
    `(Observations: ${
      yellow(new Intl.NumberFormat().format(creature.input))
    }, Neurons: ${
      yellow(
        new Intl.NumberFormat().format(
          creature.neurons.length - creature.input,
        ),
      )
    }, ` +
    `Synapses: ${
      yellow(new Intl.NumberFormat().format(creature.synapses.length))
    } & Score: ${yellow(new Intl.NumberFormat().format(startScore))})`,
);

const mutationRate = Math.min(
  0.5,
  Math.max(
    0.05, // Mutation rate of at least 5%
    args.mutationRate
      ? args.mutationRate
      : Math.random() * Math.random() * Math.random(),
  ),
);

const elitism = Math.min(
  10,
  Math.max(
    1,
    args.elitism ? args.elitism : Math.round(Math.random() * 10),
  ),
);

const mutationAmount = Math.max(
  Math.floor(10 * Math.random()),
  1,
);

let selectionName;
let selection;
if (args.selection) {
  selectionName = args.selection;
  switch (selectionName) {
    case "POWER":
      selection = Selection.POWER;
      break;
    case "TOURNAMENT":
      selection = Selection.TOURNAMENT;
      break;
    case "FITNESS_PROPORTIONATE":
      selection = Selection.FITNESS_PROPORTIONATE;
      break;
    default:
      throw "unknown selection " + selectionName;
  }
} else {
  switch (Math.floor(Math.random() * 3)) {
    case 1:
      selectionName = "POWER";
      selection = Selection.POWER;
      break;
    case 2:
      selectionName = "TOURNAMENT";
      selection = Selection.TOURNAMENT;
      break;
    default:
      selectionName = "FITNESS_PROPORTIONATE";
      selection = Selection.FITNESS_PROPORTIONATE;
  }
}

const samples = [];
const previous = [];
const creatures = [];
for (const dirEntry of Deno.readDirSync(dir)) {
  if (dirEntry.name.endsWith(".json")) {
    /** Bring in the cluster creatures. */

    const creature = await nu.load(dir + "/" + dirEntry.name);

    const regex = /[0-9][0-9]\.json/;
    if (
      dirEntry.name.match(regex)
    ) {
      previous.push(creature);
    } else {
      samples.push(creature);
    }
  }
}
/** How much each extra node subtracts from the score */
const argCostOfGrowth = Number.parseFloat(
  `${args.costOfGrowth ?? 0.000_000_1}`.replaceAll("_", ""),
);
const costOfGrowth = Math.min(
  0.000_001,
  Math.max(0.000_000_001, argCostOfGrowth),
);

const crisprList = [];
if (!args.disableCRISPR) {
  const crispr = new CRISPR(creature);
  for (const dirEntry of Deno.readDirSync("CRISPR")) {
    if (dirEntry.name.endsWith(".json")) {
      const edited = editAliases(dirEntry.name);
      try {
        const creatureCRISPR = await crispr.cleaveDNA(
          edited,
        );
        if (creatureCRISPR) {
          removeTag(creatureCRISPR, "startScore");
          crisprList.push(creatureCRISPR);
        }
      } catch (e) {
        console.error(`${dirEntry.name} CRISPR failed`, e);
      }
    }
  }
}

creatures.push(...crisprList, ...samples, ...previous);

creatures.sort((a, b) => {
  const aScoreTxt = getTag(a, "startScore");
  const aScore = aScoreTxt ? parseFloat(aScoreTxt) : -1;

  const bScoreTxt = getTag(b, "startScore");
  const bScore = bScoreTxt ? parseFloat(bScoreTxt) : -1;

  return bScore - aScore;
});

const experimentStore = ".experiments" + (args.sampler ? "/sampler" : "");
if (args.sampler) {
  emptyDirSync(experimentStore);
} else {
  ensureDirSync(experimentStore);
}
const options: NeatOptions = {
  creatureStore: dir,
  creatures: creatures,
  creativeThinkingConnectionCount: 1,
  timeoutMinutes: args.timeout ? Math.max(args.timeout, 1) : 60,
  targetError: 0.002,
  log: args.log ? Math.max(args.log, 1) : 10,
  elitism: elitism,
  mutationAmount: mutationAmount,
  mutationRate: mutationRate,
  experimentStore: experimentStore,
  selection: selection,
  mutation: mutation,
  populationSize: args.populationSize ? args.populationSize : 50,
  costOfGrowth: costOfGrowth, // The cost of growth
  threads: args.threads,
  verbose: args.verbose ? true : false,
  focusList: args.focusList ? args.focusList.split(",") : undefined,
  focusRate: args.focusRate,
  trainPerGen: args.trainPerGen,
  traceStore: args.traceStore,
  trainingSampleRate: args.trainingSampleRate,
};

console.log(
  mutationAmount,
  "mutation" + (mutationAmount > 1 ? "s" : "") +
    ", mutation rate " +
    yellow((mutationRate * 100).toFixed(1) + "%") +
    ", elitism " +
    yellow(elitism.toString()) +
    ", mutation " +
    yellow(mutationName) +
    ", selection " +
    yellow(selectionName) +
    ", timeout " +
    yellow((options.timeoutMinutes ? options.timeoutMinutes : -1).toString()) +
    ", PopSize " +
    yellow((options.populationSize ? options.populationSize : -1).toString()) +
    (args.sampler ? bold(" SAMPLER") : ""),
);

performance.mark("evolve-start");
const result = await creature.evolveDir(TRAINING_DATA_DIR, options);
performance.mark("evolve-end");

Deno.writeTextFileSync(
  ".result" + (args.sampler ? "-sampler" : "") + ".json",
  JSON.stringify(result, null, 1),
);

console.info(
  "Evolve: " +
    performance.measure("", "evolve-start", "evolve-end").duration.toFixed(
      0,
    ) + " ms",
);

const uuid = await CreatureUtil.makeUUID(creature);
faker.seed(parseInt(uuid.substring(0, 4), 16));
addTag(creature, "name", faker.name.firstName() + " " + faker.name.lastName());
addTag(creature, "version", NetworkUtil.VERSION.toString());

const json = creature.exportJSON();

await Deno.writeTextFile(
  networkFN,
  JSON.stringify(json, null, 1),
);

// try {
//   Deno.readTextFileSync(
//     ".performance" + (args.sampler ? "-sampler" : "") + ".csv",
//   );
// } catch {
//   const header =
//     "when,startScore,iterations,endScore,time,mutationAmount,mutationRate,elitism,mutation,selection,diff\n";
//   Deno.writeTextFileSync(
//     ".performance" + (args.sampler ? "-sampler" : "") + ".csv",
//     header,
//     { create: true },
//   );
// }
// const line = Date.now() + "," + startScore + "," + /*result.iterations*/ -1 +
//   "," +
//   result.score + "," +
//   result.time + "," + mutationAmount + "," + mutationRate.toFixed(2) + "," +
//   elitism + "," +
//   mutationName + "," + selectionName + "," +
//   (result.score - startScore) + "\n";

// Deno.writeTextFileSync(
//   ".performance" + (args.sampler ? "-sampler" : "") + ".csv",
//   line,
//   { append: true },
// );
